"""Contain the parsing operations to create a user indices."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from icclim._core.constants import (
    PERCENTILE_THRESHOLD_STAMP,
    USER_INDEX_PRECIPITATION_STAMP,
    USER_INDEX_TEMPERATURE_STAMP,
)
from icclim._core.legacy.user_index.calc_operation import CalcOperationRegistry
from icclim._core.model.logical_link import LogicalLink, LogicalLinkRegistry
from icclim._core.model.operator import Operator, OperatorRegistry
from icclim._core.model.threshold import Threshold
from icclim.exception import InvalidIcclimArgumentError
from icclim.generic.registry import GenericIndicatorRegistry
from icclim.threshold.factory import build_threshold

if TYPE_CHECKING:
    import datetime as dt
    from collections.abc import Sequence

    from icclim._core.generic.indicator import GenericIndicator
    from icclim._core.legacy.user_index.model import UserIndexDict
    from icclim._core.model.quantile_interpolation import QuantileInterpolation


def read_indicator(user_index: UserIndexDict) -> GenericIndicator:
    """
    Read the user index and return the corresponding generic indicator.

    Parameters
    ----------
    user_index : UserIndexDict
        The user index dictionary containing the calculation operation and extreme mode.

    Returns
    -------
    GenericIndicator
        The corresponding generic indicator based on the user index.

    Raises
    ------
    InvalidIcclimArgumentError
        If the user index does not contain a calculation operation.
        If the user index's calculation operation is unknown.
    NotImplementedError
        If the calculation operation or extreme mode is not implemented.

    Notes
    -----
    This function reads the user index dictionary and maps the calculation operation and
    extreme mode to the corresponding generic indicator.
    It raises errors if the required information is missing or if the operation is not
    implemented.
    """
    calc_op = user_index.get("calc_operation", None)
    if calc_op is None:
        err = "user_index needs a calc_operation"
        raise InvalidIcclimArgumentError(err)
    calc_op = CalcOperationRegistry.lookup(calc_op)
    user_index_map = {
        CalcOperationRegistry.MAX: GenericIndicatorRegistry.Maximum.clone(),
        CalcOperationRegistry.MIN: GenericIndicatorRegistry.Minimum.clone(),
        CalcOperationRegistry.SUM: GenericIndicatorRegistry.Sum.clone(),
        CalcOperationRegistry.MEAN: GenericIndicatorRegistry.Average.clone(),
        CalcOperationRegistry.EVENT_COUNT: GenericIndicatorRegistry.CountOccurrences.clone(),  # noqa: E501
        CalcOperationRegistry.MAX_NUMBER_OF_CONSECUTIVE_EVENTS: GenericIndicatorRegistry.MaxConsecutiveOccurrence.clone(),  # noqa: E501
        CalcOperationRegistry.ANOMALY: GenericIndicatorRegistry.DifferenceOfMeans.clone(),  # noqa: E501
    }
    extrem_mode = user_index.get("extreme_mode", None)
    if calc_op == CalcOperationRegistry.RUN_SUM:
        if extrem_mode == "max":
            indicator = GenericIndicatorRegistry.MaxOfRollingSum.clone()
        elif extrem_mode == "min":
            indicator = GenericIndicatorRegistry.MinOfRollingSum.clone()
        else:
            raise NotImplementedError
    elif calc_op == CalcOperationRegistry.RUN_MEAN:
        if extrem_mode == "max":
            indicator = GenericIndicatorRegistry.MaxOfRollingAverage.clone()
        elif extrem_mode == "min":
            indicator = GenericIndicatorRegistry.MinOfRollingAverage.clone()
        else:
            raise NotImplementedError
    else:
        indicator = user_index_map.get(calc_op)
    if indicator is None:
        msg = (
            "Unknown user_index's calc_operation:"
            f" '{user_index.get('calc_operation', None)}'"
        )
        raise InvalidIcclimArgumentError(msg)
    return indicator


def read_logical_link(user_index: UserIndexDict) -> LogicalLink:
    """
    Read the logical link from the user index dictionary.

    Parameters
    ----------
    user_index : UserIndexDict
        The user index dictionary containing the logical link information.

    Returns
    -------
    LogicalLink
        The corresponding LogicalLink based on the logical link information in the user
        index dictionary.

    Notes
    -----
    If the logical link is not specified in the user index dictionary, the default
    logical link is LogicalLinkRegistry.LOGICAL_AND.
    """
    logical_link = user_index.get("link_logical_operations", None)
    if logical_link is None:
        return LogicalLinkRegistry.LOGICAL_AND
    return LogicalLinkRegistry.lookup(logical_link)


def read_coef(user_index: UserIndexDict) -> float | None:
    """
    Read the coefficient value from the user index dictionary.

    Parameters
    ----------
    user_index : UserIndexDict
        The user index dictionary containing the coefficient value.

    Returns
    -------
    float or None
        The coefficient value if it exists in the user index dictionary, otherwise None.
    """
    return user_index.get("coef", None)


def read_date_event(user_index: UserIndexDict) -> bool:
    """
    Read the 'date_event' key from the given UserIndexDict.

    Parameters
    ----------
    user_index : UserIndexDict
        The dictionary containing user index information.

    Returns
    -------
    bool
        The value associated with the 'date_event' key in the UserIndexDict,
        if missing returns False.
    """
    return user_index.get("date_event", False)


def read_thresholds(
    user_index: UserIndexDict,
    doy_window_width: int,
    reference_period: Sequence[dt.datetime | str] | None,
    only_leap_years: bool,
    interpolation: QuantileInterpolation,
) -> Threshold | None | list[Threshold]:
    """
    Read the thresholds from the user index dictionary.

    Parameters
    ----------
    user_index : UserIndexDict
        The user index dictionary containing the threshold information.
    doy_window_width : int
        The width of the day of year window for calculating the threshold.
    reference_period : Sequence[dt.datetime | str] | None
        The reference period for calculating the threshold.
    only_leap_years : bool
        Whether to consider only leap years when calculating the threshold.
    interpolation : QuantileInterpolation
        The interpolation method to use for calculating the threshold.

    Returns
    -------
    Threshold or None or list[Threshold]
        The corresponding Threshold object(s) based on the threshold information in the
        user index dictionary.

    Notes
    -----
    This function reads the threshold information from the user index dictionary and
    maps it to the corresponding Threshold object(s).
    If the threshold is already a Threshold object, it is returned as is.
    If the threshold is a tuple or list, multiple Threshold objects are created based
    on the logical operation and link specified in the user index dictionary.
    If the threshold is a single value, a single Threshold object is created.
    """
    thresh = user_index.get("thresh", None)
    var_type = user_index.get("var_type", None)
    if thresh is None or isinstance(thresh, Threshold):
        return thresh
    logical_operation = user_index.get("logical_operation", None)
    if isinstance(thresh, (tuple, list)):
        if logical_operation is None:
            logical_operations = [OperatorRegistry.REACH for _ in thresh]
        else:
            logical_operations = [
                OperatorRegistry.lookup(lo) for lo in logical_operation
            ]
        link = user_index.get(
            "link_logical_operations", LogicalLinkRegistry.LOGICAL_AND
        )
        link = LogicalLinkRegistry.LOGICAL_AND if link is None else link
        link = LogicalLinkRegistry.lookup(link)
        thresholds = (
            _build_thresh_query(t, var_type, logical_operations[i])
            for i, t in enumerate(thresh)
        )
        return [
            build_threshold(
                t,
                doy_window_width=doy_window_width,
                reference_period=reference_period,
                only_leap_years=only_leap_years,
                interpolation=interpolation,
            )
            for t in thresholds
        ]
    if logical_operation is None:
        logical_operation = OperatorRegistry.REACH
    elif isinstance(logical_operation, str):
        logical_operation = OperatorRegistry.lookup(logical_operation)
    elif isinstance(logical_operation, Operator):
        pass
    else:
        err = "Invalid user_index: logical_operation must be None or a string."
        raise InvalidIcclimArgumentError(err)
    thresh_query = _build_thresh_query(thresh, var_type, logical_operation)
    return build_threshold(
        thresh_query,
        doy_window_width=doy_window_width,
        reference_period=reference_period,
        only_leap_years=only_leap_years,
        interpolation=interpolation,
    )


def _build_thresh_query(
    query: str | float,
    var_type: Literal["t", "p"] | None,
    operator: Operator,
) -> str:
    if isinstance(query, str) and query.endswith(PERCENTILE_THRESHOLD_STAMP):
        if var_type == USER_INDEX_TEMPERATURE_STAMP:
            replace_unit = "doy_per"
        elif var_type == USER_INDEX_PRECIPITATION_STAMP:
            replace_unit = "period_per"
        else:
            replace_unit = "period_per"  # default to period percentiles ?
        t = query.replace(PERCENTILE_THRESHOLD_STAMP, " " + replace_unit)
    else:
        t = str(query)
    return f"{operator.operand} {t}"



"""Contain the registry of the DCSC (Meteo France) specific indices."""

from __future__ import annotations

from icclim._core.constants import NEEDS_NORMAL, QUANTILE_BASED, REFERENCE_PERIOD_INDEX
from icclim._core.model.index_group import IndexGroupRegistry
from icclim._core.model.registry import Registry
from icclim._core.model.standard_index import StandardIndex
from icclim._core.model.standard_variable import StandardVariableRegistry
from icclim.generic.registry import (
    GenericIndicatorRegistry,
)
from icclim.threshold.factory import build_threshold

DCSC_REFERENCE = "Portail DRIAS, DCSC, MeteoFrance"
DCSC_SOURCE = "Portail DRIAS, DCSC, MeteoFrance"


class DcscIndexRegistry(Registry[StandardIndex]):
    """
    Registry of the DCSC (Meteo France) specific indices.

    Note
    ----

    The indices metadata of this module are in French.
    """

    _item_class = StandardIndex

    @staticmethod
    def get_item_aliases(item: StandardIndex) -> list[str]:
        """
        Duck-typed method to get the aliases of a StandardIndex item.

        Parameters
        ----------
        item : StandardIndex
            The StandardIndex item.

        Returns
        -------
        list[str]
            The aliases of the item.

        Notes
        -----
        Every StandardIndex registry should implement this method.
        """
        return [item.short_name]

    TAV = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="degree_Celsius",
        definition="Moyenne de la température moyenne.",
        source=DCSC_SOURCE,
        short_name="TAV",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[StandardVariableRegistry.TAS],
    )
    TXAV = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="degree_Celsius",
        definition="Moyenne de la température maximale.",
        source=DCSC_SOURCE,
        short_name="TXAV",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[StandardVariableRegistry.TAS_MAX],
    )
    TRAV = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.MeanOfDifference,
        output_unit="degree_Celsius",
        definition="Moyenne de l'amplitude thermique.",
        source=DCSC_SOURCE,
        short_name="TRAV",
        group=IndexGroupRegistry.TEMPERATURE,
        input_variables=[
            StandardVariableRegistry.TAS_MIN,
            StandardVariableRegistry.TAS_MAX,
        ],
    )
    TX10 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 10 doy_per",
        output_unit="day",
        definition=(
            "Extrême froid de la température maximale"
            " journalière (10e centile de la température"
            " maximale)."
        ),
        source=DCSC_SOURCE,
        short_name="TX10",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TX90 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="> 90 doy_per",
        output_unit="day",
        definition=(
            "Extrême chaud de la température maximale journalière"
            " (90e centile de la température maximale)."
        ),
        source=DCSC_SOURCE,
        short_name="TX90",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TN10 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 10 doy_per",
        output_unit="day",
        definition=(
            "Extrême froid de la température minimale "
            " journalière (10e centile de la température minimale)."
        ),
        source=DCSC_SOURCE,
        short_name="TN10",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TN90 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition=(
            "Extrême chaud de la température minimale"
            " journalière (90e centile de la température minimale)."
        ),
        threshold="> 90 doy_per",
        source=DCSC_SOURCE,
        short_name="TN90",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[QUANTILE_BASED],
        doy_window_width=5,
    )
    TNFD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 0 degree_Celsius",
        output_unit="day",
        definition="Nombre de jours de gel (température minimale <= 0°C).",
        source=DCSC_SOURCE,
        short_name="TNFD",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[],
    )
    TXFD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold="< 0 degree_Celsius",
        output_unit="day",
        definition="Nombre de jours sans dégel (température maximale <= 0°C).",
        source=DCSC_SOURCE,
        short_name="TXFD",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[],
    )
    SD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition="Nombre de journées d'été (température maximale > 25°C).",
        source=DCSC_SOURCE,
        short_name="SD",
        threshold="> 25 degree_Celsius",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[],
    )
    TX35 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition="Nombre de jours de forte chaleur (température maximale > 35°C).",
        source=DCSC_SOURCE,
        short_name="TX35",
        threshold="> 35 degree_Celsius",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[],
    )
    TR = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition="Nombre de nuits tropicales (température minimale > 20°C).",
        source=DCSC_SOURCE,
        short_name="TR",
        threshold="> 20 degree_Celsius",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[],
    )
    TXND = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        definition=(
            "Nombre de jours anormalement chauds"
            " (température maximale supérieure de plus de 5°C à la normale)."
        ),
        source=DCSC_SOURCE,
        short_name="TXND",
        threshold=build_threshold(
            operator=">",
            value=None,  # filled when Threshold::prepare is called
            unit="degree_Celsius",
            offset=" 5 delta_degree_Celsius",
        ),
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[NEEDS_NORMAL],
    )
    TNHT = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        threshold=build_threshold(
            operator=">",
            value=None,  # filled when Threshold::prepare is called
            unit="degree_Celsius",
            offset=" 5 delta_degree_Celsius",
        ),
        source=DCSC_SOURCE,
        short_name="TNHT",
        definition="Nombre de nuits anormalement chaudes (température minimale "
        "supérieure de plus de 5°C à la normale).",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[NEEDS_NORMAL],
    )
    TNND = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        output_unit="day",
        threshold=build_threshold(
            operator="<",
            value=None,  # filled when Threshold::prepare is called
            unit="degree_Celsius",
            offset=" 5 delta_degree_Celsius",
        ),
        source=DCSC_SOURCE,
        short_name="TNND",
        definition="Nombre de jours anormalement froids"
        " (température minimale inférieure de plus de 5°C à la normale).",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[NEEDS_NORMAL],
    )
    TNCWD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.SumOfSpellLengths,
        output_unit="day",
        threshold=build_threshold(
            operator="<",
            value=None,  # filled when Threshold::prepare is called
            unit="degree_Celsius",
            offset=" 5 delta_degree_Celsius",
        ),
        source=DCSC_SOURCE,
        short_name="TNCWD",
        definition="Nombre de jours d'une vague de froid"
        " (température min < de plus de 5°C à la normale pdt au moins "
        "5j consécutifs).",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS_MIN],
        qualifiers=[NEEDS_NORMAL],
    )
    TXHWD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.SumOfSpellLengths,
        output_unit="day",
        threshold=build_threshold(
            operator="<",
            value=None,  # filled when Threshold::prepare is called
            unit="degree_Celsius",
            offset=" 5 delta_degree_Celsius",
        ),
        source=DCSC_SOURCE,
        short_name="TXHWD",
        definition="Nombre de jours d'une vague de chaleur"
        " (température max > de plus de 5°C à la normale"
        " pdt au moins 5j consécutifs).",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS_MAX],
        qualifiers=[NEEDS_NORMAL],
    )
    HDD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Deficit,
        threshold="17 degree_Celsius",
        output_unit="degree_Celsius day",
        definition="Degrés-jours de chauffage"
        " (Cumul sur la période des écarts négatifs au seuil de < 17°C"
        " par la température qt moyenne).",
        source=DCSC_SOURCE,
        short_name="HDD",
        group=IndexGroupRegistry.COLD,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[],
    )
    CDD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Excess,
        threshold="18 degree_Celsius",
        output_unit="degree_Celsius day",
        definition="Degrés-jours de climatisation"
        "(Cumul sur la période des dépassements du seuil de > 18°C"
        " par la température qt moyenne).",
        source=DCSC_SOURCE,
        short_name="CDD",
        group=IndexGroupRegistry.HEAT,
        input_variables=[StandardVariableRegistry.TAS],
        qualifiers=[],
    )
    # PRECIPITATION
    PAV = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        output_unit="mm/day",
        definition="Précipitations quotidiennes moyennes.",
        source=DCSC_SOURCE,
        short_name="PAV",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    PINT = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Average,
        threshold=">= 1 mm/day",
        output_unit="mm/day",
        definition="Précipitation moyenne des jours pluvieux (RR > 1 mm).",
        source=DCSC_SOURCE,
        short_name="PINT",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    RR = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Sum,
        output_unit="mm",
        definition="Cumul de précipitation.",
        source=DCSC_SOURCE,
        short_name="RR",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    RR1MM = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 1 mm/day",
        output_unit="day",
        definition="Nombre de jours de pluie (précipitations >= 1 mm).",
        source=DCSC_SOURCE,
        short_name="RR1MM",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    PN20MM = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=">= 20 mm/day",
        output_unit="day",
        definition="Nombre de jours de fortes précipitations"
        " (précipitations >= 20 mm).",
        source=DCSC_SOURCE,
        short_name="PN20MM",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    PXCDD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.MaxConsecutiveOccurrence,
        threshold="< 1 mm/day",
        output_unit="day",
        definition="Période de sécheresse (Max [Nbj consécutifs RR < 1 mm]).",
        source=DCSC_SOURCE,
        short_name="PXCDD",
        group=IndexGroupRegistry.DROUGHT,
        input_variables=[StandardVariableRegistry.PR],
    )
    PXCWD = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.MaxConsecutiveOccurrence,
        threshold=">= 1 mm/day",
        output_unit="day",
        definition="Nombre maximum de jours pluvieux consécutifs"
        " (Max [Nbj consécutifs RR > 1 mm]).",
        source=DCSC_SOURCE,
        short_name="PXCWD",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
    )
    R99 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=build_threshold("> 99 period_per", threshold_min_value="1 mm/day"),
        output_unit="day",
        definition="Nombre de jours de précipitations extrêmes.",
        source=DCSC_SOURCE,
        short_name="R99",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    PFL90 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.FractionOfTotal,
        threshold=build_threshold("> 90 period_per", threshold_min_value="1 mm/day"),
        output_unit="%",
        definition="Fraction des précipitations journalières intenses.",
        source=DCSC_SOURCE,
        short_name="PFL90",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    PQ90 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Percentile,
        threshold=build_threshold("> 90 doy_per", threshold_min_value="1 mm/day"),
        output_unit="%",
        definition="Précipitation quotidienne intense"
        " (90e centile des précipitations).",
        source=DCSC_SOURCE,
        short_name="PQ90",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    PQ99 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.Percentile,
        threshold=build_threshold("> 99 doy_per", threshold_min_value="1 mm/day"),
        output_unit="%",
        definition="Précipitation quotidienne extrême"
        " (99e centile des précipitations).",
        source=DCSC_SOURCE,
        short_name="PQ99",
        group=IndexGroupRegistry.RAIN,
        input_variables=[StandardVariableRegistry.PR],
        qualifiers=[QUANTILE_BASED],
    )
    # VITESSE DE VENT
    FFAV = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.DifferenceOfMeans,
        output_unit="m s-1",
        definition="Écart de la vitesse du vent moyenne journalière"
        " (par rapport à une periode de référence).",
        source=DCSC_SOURCE,
        short_name="FFAV",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.SFC_WIND],
        qualifiers=[REFERENCE_PERIOD_INDEX],
    )
    FF98 = StandardIndex(
        reference=DCSC_REFERENCE,
        indicator=GenericIndicatorRegistry.CountOccurrences,
        threshold=build_threshold("> 98 period_per", threshold_min_value="1 knots"),
        output_unit="days",
        definition="Nombre de jours de vent fort"
        " (vent ≥ 98e centile de la période de référence).",
        source=DCSC_SOURCE,
        short_name="FF98",
        group=IndexGroupRegistry.WIND,
        input_variables=[StandardVariableRegistry.SFC_WIND],
        qualifiers=[REFERENCE_PERIOD_INDEX, QUANTILE_BASED],
    )



"""
Contain the ClimateVariable class and its related functions.

A climate variable is a structure that contains all the pre-processed input varaible to
compute a climate index.
A climate index may require one or more climate variables to be computed.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

import numpy as np
import xarray

from icclim._core.constants import REFERENCE_PERIOD_INDEX, UNITS_KEY
from icclim._core.generic.threshold.percentile import PercentileThreshold
from icclim._core.input_parsing import (
    DEFAULT_INPUT_FREQUENCY,
    build_reference_da,
    build_studied_data,
    guess_standard_variable,
    read_dataset,
)
from icclim.exception import InvalidIcclimArgumentError
from icclim.frequency import Frequency, FrequencyRegistry
from icclim.threshold.factory import build_threshold

if TYPE_CHECKING:
    from collections.abc import Sequence
    from datetime import datetime

    import jinja2
    from xarray.core.dataarray import DataArray

    from icclim._core.model.global_metadata import GlobalMetadata
    from icclim._core.model.icclim_types import InFileBaseType
    from icclim._core.model.in_file_dictionary import InFileDictionary
    from icclim._core.model.standard_index import StandardIndex
    from icclim._core.model.standard_variable import StandardVariable
    from icclim._core.model.threshold import Threshold


@dataclass
class ClimateVariable:
    """
    ClimateVariable is a dataclass that represents a climate variable used to compute a climate index.

    It groups together the input variable (studied_data), its associated metadata
    (standard_var) if any, the threshold it must be compared to.

    Attributes
    ----------
    name: str
        Name of the variable.
    standard_var: StandardVariable
        CF metadata bounded to the standard variable used for this ClimateVariable.
    studied_data: DataArray
        The variable studied.
    threshold: Threshold | None
        thresholds for this variable
    reference_period: Sequence of str | None
        The reference period to consider
    """  # noqa: E501

    name: str
    standard_var: StandardVariable | None
    studied_data: DataArray
    global_metadata: GlobalMetadata
    source_frequency: Frequency
    threshold: Threshold | None = None
    reference_period: Sequence[datetime | str] | None = None
    is_reference: bool = False

    def build_indicator_metadata(
        self,
        src_freq: Frequency,
        must_run_bootstrap: bool,
        jinja_scope: dict[str, Any],
        jinja_env: jinja2.Environment,
    ) -> dict[str, str | dict]:
        """
        Build the metadata for the indicator that will be computed with this variable.

        Parameters
        ----------
        src_freq: Frequency
            The frequency of the source data.
        must_run_bootstrap: bool
            Whether the bootstrap method must be run.
        jinja_scope: dict
            The scope to use for jinja templating.
        jinja_env: jinja2.Environment
            The environment to use for jinja templating.

        Returns
        -------
        dict of str, str | dict
            The metadata for the indicator.
        """
        metadata: dict[str, str | dict] = {"threshold": {}}
        if self.standard_var is None:
            metadata.update(
                {
                    "standard_name": "unknown_variable",
                    "long_name": "unknown variable",
                    "short_name": "input",
                },
            )
        else:
            metadata.update(self.standard_var.get_metadata())
        if self.threshold is not None:
            metadata.update(
                {
                    "threshold": self.threshold.format_metadata(
                        src_freq=src_freq,
                        must_run_bootstrap=must_run_bootstrap,
                        jinja_scope=jinja_scope,
                        jinja_env=jinja_env,
                    ),
                },
            )
        return metadata


def build_climate_vars(
    climate_vars_dict: dict[str, InFileDictionary],
    ignore_Feb29th: bool,  # noqa: N803
    time_range: Sequence[datetime | str] | None,
    base_period: Sequence[str] | None,
    standard_index: StandardIndex | None,
    is_compared_to_reference: bool,
) -> list[ClimateVariable]:
    """
    Build a list of ClimateVariable from a dictionary of input files.

    Parameters
    ----------
    climate_vars_dict: dict of str, InFileDictionary
        The dictionary of input files.
    ignore_Feb29th: bool
        Whether to ignore February 29th.
    time_range: Sequence of datetime | str | None
        The time range to consider.
    base_period: Sequence of str | None
        The base period to consider, used to build a reference variable for indices such
         as anomaly.
    standard_index: StandardIndex | None
        The standard index to compute.

    Returns
    -------
    list of ClimateVariable that will be used to compute the climate index.
    """
    from icclim.ecad.binding import (
        StandardizedPrecipitationIndex3,
        StandardizedPrecipitationIndex6,
    )

    if standard_index is not None and len(standard_index.input_variables) > len(
        climate_vars_dict
    ):
        msg = (
            f"Index {standard_index.short_name} needs"
            f" {len(standard_index.input_variables)} variables."
            f" Please provide them with an xarray.Dataset, netCDF file(s) or a"
            f" zarr store."
        )
        raise InvalidIcclimArgumentError(msg)
    acc = []
    for i, raw_climate_var in enumerate(climate_vars_dict.items()):
        if standard_index is not None:
            standard_var = standard_index.input_variables[i]
        else:
            standard_var = None

        # For SPI, attach reference_period directly to the study variable
        if standard_index is not None and standard_index.short_name.lower() in (
            "spi3",
            "spi6",
        ):
            reference_period = base_period
        else:
            reference_period = None

        cv = build_climate_var(
            raw_climate_var[0],
            raw_climate_var[1],
            ignore_Feb29th,
            time_range,
            standard_var=standard_var,
            reference_period=reference_period,
        )

        acc.append(cv)

    # Only add a reference variable for non-SPI indices
    if not isinstance(
        standard_index,
        (StandardizedPrecipitationIndex3, StandardizedPrecipitationIndex6),
    ):
        if _standard_index_needs_ref(
            standard_index, is_compared_to_reference
        ) or _generic_index_needs_ref(standard_index, is_compared_to_reference):
            standard_var = standard_index.input_variables[0] if standard_index else None
            added_var = _build_reference_variable(
                base_period,
                climate_vars_dict,
                standard_var=standard_var,
            )
            acc.append(added_var)

    return acc


def build_climate_var(
    climate_var_name: str,
    climate_var_data: InFileDictionary | InFileBaseType,
    ignore_Feb29th: bool,  # noqa: N803
    time_range: Sequence[datetime | str] | None,
    standard_var: StandardVariable | None,
    reference_period: Sequence[datetime | str] | None = None,
) -> ClimateVariable:
    """
    Build a ClimateVariable object.

    Parameters
    ----------
    climate_var_name : str
        The name of the climate variable.
    climate_var_data : InFileDictionary | InFileBaseType
        The input data for the climate variable. It can be either a dictionary
        or a file.
    ignore_Feb29th : bool
        Flag indicating whether to ignore February 29th in the time range.
    time_range : Sequence[datetime | str] | None
        The time range to consider for the climate variable. It can be a sequence
        of datetime objects or strings, or None to consider the entire time range.
    standard_var : StandardVariable | None
        The standard variable to use for the climate variable. If None, the input
        data will be used to guess the standard variable.

    Returns
    -------
    ClimateVariable
        The built ClimateVariable object.

    Notes
    -----
    This function builds a ClimateVariable object based on the provided inputs.
    It reads the input data, determines the standard variable, builds the studied
    data, and sets the threshold and global metadata.

    If the input data is a dictionary, it is assumed to have a 'study' key
    containing the study data and an optional 'thresholds' key containing the
    threshold data.

    If the input data is a file, it is assumed to contain the study data.

    The standard variable is used to determine the conversion unit for the
    threshold data.

    The studied data is built based on the study data, time range, ignore_Feb29th
    flag, and standard variable.

    If a threshold is provided in the dictionary, it is added to the ClimateVariable.

    Examples
    --------
    >>> climate_var_name = "tas"
    >>> climate_var_data = {"study": "/path/to/data.nc", "thresholds": ">= 27 degC"}
    >>> ignore_Feb29th = False
    >>> time_range = ["2000-01-01", "2010-12-31"]
    >>> standard_var = StandardVariableRegistry.TAS
    >>> climate_var = build_climate_var(
    ...     climate_var_name, climate_var_data, ignore_Feb29th, time_range, standard_var
    ... )
    """
    if isinstance(climate_var_data, dict):
        study_ds = read_dataset(
            climate_var_data["study"],
            standard_var,
            climate_var_name,
        )
        climate_var_thresh = climate_var_data.get("thresholds", None)
    else:
        study_ds = read_dataset(climate_var_data, standard_var, climate_var_name)
        climate_var_thresh = None
    if standard_var is None:
        standard_var = guess_standard_variable(study_ds[climate_var_name])
    studied_data = build_studied_data(
        study_ds[climate_var_name],
        time_range,
        ignore_Feb29th,
        standard_var.default_units if standard_var else None,
    )
    if climate_var_thresh is not None:
        climate_var_thresh = _build_threshold(
            climate_var_thresh=climate_var_thresh,
            original_data=study_ds[climate_var_name],
            conversion_unit=studied_data.attrs[UNITS_KEY],
        )
    return ClimateVariable(
        name=climate_var_name,
        standard_var=standard_var,
        studied_data=studied_data,
        threshold=climate_var_thresh,
        reference_period=reference_period,
        global_metadata={
            "history": study_ds.attrs.get("history", None),
            "source": study_ds.attrs.get("source", None),
            "time_encoding": study_ds.time.encoding,
        },
        source_frequency=FrequencyRegistry.lookup(
            xarray.infer_freq(studied_data.time) or DEFAULT_INPUT_FREQUENCY,
        ),
    )


def must_run_bootstrap(da: DataArray, threshold: Threshold | None) -> bool:
    """
    Determine whether to run the bootstrap method.

    Parameters
    ----------
    da : DataArray
        The studied data.
    threshold : Threshold | None
        The threshold that contains the reference period.

    Returns
    -------
    bool
        Whether to run the bootstrap method.

    Notes
    -----
    This function is used to avoid bootstrapping if there is one single year
    overlapping or no year overlapping or all year overlapping between the studied
    data `da` and the reference period defined by the threshold.
    """
    # TODO @bzah: Don't run bootstrap when not on extreme percentile
    #       (run only below 20? 10? and above 80? 90?)
    # https://github.com/cerfacs-globc/icclim/issues/289
    if (
        threshold is None
        or not isinstance(threshold, PercentileThreshold)
        or (
            isinstance(threshold, PercentileThreshold)
            and not threshold.is_doy_per_threshold
        )
    ):
        return False
    reference = threshold.value
    study_years = np.unique(da.indexes.get("time").year)
    overlapping_years = np.unique(
        da.sel(time=_get_ref_period_slice(reference)).indexes.get("time").year,
    )
    return 1 < len(overlapping_years) < len(study_years)


def _standard_index_needs_ref(
    standard_index: StandardIndex, is_compared_to_reference: bool
) -> bool:
    return (
        standard_index
        and standard_index.qualifiers
        and REFERENCE_PERIOD_INDEX in standard_index.qualifiers
        and is_compared_to_reference
    )


def _generic_index_needs_ref(
    standard_index: StandardIndex, is_compared_to_reference: bool
) -> bool:
    return standard_index is None and is_compared_to_reference


def _build_reference_variable(
    reference_period: Sequence[str] | None,
    in_files: dict[str, InFileDictionary],
    standard_var: StandardVariable,
) -> ClimateVariable:
    """
    Add a secondary variable for indices such as anomaly.

    This kind of indices require exactly two variables, but the second variable can
    just be a subset of the first one.
    """
    if reference_period is None:
        msg = "Can't build a reference variable without a `base_period_time_range`"
        raise InvalidIcclimArgumentError(msg)
    var_name = next(iter(in_files.keys()))
    if isinstance(in_files, dict):
        study_ds = read_dataset(
            next(iter(in_files.values()))["study"],
            standard_var=standard_var,
            var_name=var_name,
        )
    else:
        study_ds = read_dataset(
            next(iter(in_files.values())),
            standard_var=standard_var,
            var_name=var_name,
        )
    studied_data = build_reference_da(
        study_ds[var_name],
        reference_period,
        only_leap_years=False,
        percentile_min_value=None,
    )
    return ClimateVariable(
        name=var_name + "_reference",
        standard_var=standard_var,
        studied_data=studied_data,
        threshold=None,
        reference_period=reference_period,
        global_metadata={
            "history": study_ds.attrs.get("history", None),
            "source": study_ds.attrs.get("source", None),
            "time_encoding": study_ds.time.encoding,
        },
        source_frequency=FrequencyRegistry.lookup(
            xarray.infer_freq(studied_data.time) or DEFAULT_INPUT_FREQUENCY,
        ),
        is_reference=True,
    )


def _build_threshold(
    climate_var_thresh: str | Threshold,
    original_data: DataArray,
    conversion_unit: str,
) -> Threshold:
    if isinstance(climate_var_thresh, str):
        climate_var_thresh: Threshold = build_threshold(climate_var_thresh)
    if climate_var_thresh.prepare is not None and not climate_var_thresh.is_ready:
        climate_var_thresh.prepare(original_data)
    climate_var_thresh.unit = conversion_unit
    return climate_var_thresh


def _get_ref_period_slice(da: DataArray) -> slice:
    if (bds := da.attrs.get("climatology_bounds", None)) is not None:
        return slice(*bds)
    time_length = len(da.time)
    return slice(*da.time[0 :: time_length - 1].dt.strftime("%Y-%m-%d").to_numpy())



"""Threshold abstract class and ThresholdBuilderInput type."""

from __future__ import annotations

import abc
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, TypedDict, Union

import xarray as xr
from xarray import DataArray, Dataset

from icclim._core.model.logical_link import LogicalLink, LogicalLinkRegistry

if TYPE_CHECKING:
    from datetime import datetime

    import jinja2
    import pint

    from icclim._core.generic.threshold.bounded import BoundedThreshold
    from icclim._core.generic.threshold.threshold_templates import (
        ThresholdMetadata,
    )
    from icclim._core.model.operator import Operator
    from icclim._core.model.quantile_interpolation import (
        QuantileInterpolation,
    )


ThresholdValueType = Union[
    str,
    float,
    int,
    Dataset,
    DataArray,
    Sequence[float | int | str],
    None,
]


class ThresholdBuilderInput(TypedDict, total=False):
    """Threshold building configuration.

    Data transfert object mapping all possible configuration to build any threshold.
    """

    operator: Operator
    value: ThresholdValueType
    unit: str | None
    threshold_var_name: str | None
    initial_query: str | None
    threshold_min_value: pint.Quantity | None
    offset: pint.Quantity | None
    # percentile conf:
    doy_window_width: int | None
    only_leap_years: bool | None
    interpolation: str | QuantileInterpolation | None
    reference_period: Sequence[datetime | str] | None
    # bounded conf:
    thresholds: (
        tuple[
            ThresholdBuilderInput | Threshold,
            ThresholdBuilderInput | Threshold,
        ]
        | None
    )
    logical_link: LogicalLink


class Threshold(abc.ABC):
    """
    Abstract class for all thresholds.

    See :ref:`generic_indices_recipes` for how to use custom thresholds.
    """

    operator: Operator | str
    value: ThresholdValueType
    unit: str | None = None
    initial_query: str | None
    threshold_min_value: pint.Quantity | None = None
    threshold_var_name: str | None
    prepare: Callable | None = None
    is_ready: bool | None = None

    @abc.abstractmethod
    def format_metadata(
        self,
        *,
        jinja_scope: dict[str, Any],
        jinja_env: jinja2.Environment,
        **kwargs,
    ) -> ThresholdMetadata:
        """Get a dictionary of standardized threshold metadata."""
        ...

    @abc.abstractmethod
    def __eq__(self, other: Threshold) -> bool:
        """Check if two Threshold are equal."""
        ...

    def __and__(self, other: Threshold) -> BoundedThreshold:
        """Build a BoundedThreshold from two existing Threshold with a "AND" LogicalLink."""  # noqa: E501
        from icclim._core.generic.threshold.bounded import BoundedThreshold

        return BoundedThreshold(
            thresholds=[self, other],
            logical_link=LogicalLinkRegistry.LOGICAL_AND,
            initial_query=None,
        )

    def __or__(self, other: Threshold) -> BoundedThreshold:
        """Build a BoundedThreshold from two existing Threshold with a "OR" LogicalLink."""  # noqa: E501
        from icclim._core.generic.threshold.bounded import BoundedThreshold

        return BoundedThreshold(
            thresholds=[self, other],
            logical_link=LogicalLinkRegistry.LOGICAL_OR,
            initial_query=None,
        )

    @abc.abstractmethod
    def compute(
        self,
        comparison_data: xr.DataArray,
        override_op: Callable[[DataArray, DataArray], DataArray] | None = None,
        **kwargs,
    ) -> DataArray:
        """
        Compute the exceedance of the threshold.

        For example, if the threshold is 10 and the comparison_data is [5, 10, 15],
        with a ">" operator, the exceedance will be [False, False, True].
        The operator can be overridden by `override_op`.
        This is needed when self.operator is REACH.

        Parameters
        ----------
        comparison_data: xr.DataArray
            Data that must be compared to self threshold
        override_op: Callable[[DataArray, DataArray], DataArray] | None
            Operator to override self.operator compute function.
            Optional.
        kwargs:
            Keyword arguments passed to the concrete `compute` implementations.
            This makes the `compute` interface contract unreliable.
            So we accept to not respected LSP here.
        """
        ...



"""Direction de la Climatologie et des Services Climatiques (DCSC) indices.

Meteo France's Direction de la Climatologie et des Services Climatiques (DCSC)
specialized indices.

The DCSC indices public API, via the `icclim.dcsc` package, is generated from the
`icclim.dcsc.registry.DcscIndexRegistry` registry definitions.
The parameters  of the functions are specialized to each index but are all taken from
`icclim.main.index` general function.
In other words, the DCSC indices in `icclim.dcsc` module are specializations of
`icclim.main.index` for DCSC indices.
"""

from icclim._generated._dcsc import *  # noqa: F403



"""Contain the GenericIndicator class."""

from __future__ import annotations

import contextlib
from collections.abc import Callable
from copy import deepcopy
from functools import reduce
from typing import TYPE_CHECKING, Any

import numpy as np
import xarray as xr
from jinja2 import Environment
from xarray import DataArray
from xclim.core.calendar import select_time
from xclim.core.cfchecks import cfcheck_from_name
from xclim.core.missing import MISSING_METHODS
from xclim.core.units import convert_units_to, rate2amount
from xclim.core.units import units as ureg

try:
    from pint import (
        DefinitionSyntaxError,
        DimensionalityError,
        OffsetUnitCalculusError,
        UndefinedUnitError,
    )
except ImportError:
    pass


from icclim._core.climate_variable import must_run_bootstrap
from icclim._core.constants import (
    RESAMPLE_METHOD,
)
from icclim._core.generic.functions import check_freq
from icclim._core.generic.generic_templates import INDICATORS_TEMPLATES_EN
from icclim._core.model.indicator import Indicator
from icclim.exception import InvalidIcclimArgumentError

if TYPE_CHECKING:
    import jinja2

    from icclim._core.climate_variable import ClimateVariable
    from icclim._core.model.index_config import IndexConfig
    from icclim.frequency import Frequency


# Hydro context for precipitation-like variables
context_hydro = ureg.Context("hydro")
context_hydro.add_transformation(
    "[mass] / [length] ** 2",
    "[length]",
    lambda ureg, x: x * ureg("1 mm") / ureg("1 kg / m^2"),
)
context_hydro.add_transformation(
    "[length]",
    "[mass] / [length] ** 2",
    lambda ureg, x: x * ureg("1 kg / m^2") / ureg("1 mm"),
)
context_hydro.add_transformation(
    "[mass] / [length] ** 2 / [time]",
    "[length] / [time]",
    lambda ureg, x: x * ureg("1 mm / s") / ureg("1 kg / m^2 / s"),
)
context_hydro.add_transformation(
    "[length] / [time]",
    "[mass] / [length] ** 2 / [time]",
    lambda ureg, x: x * ureg("1 kg / m^2 / s") / ureg("1 mm / s"),
)


jinja_env = Environment(autoescape=True)


class GenericIndicator(Indicator):
    """
    GenericIndicator are climate indicators wich are not specific to a particular domain.

    They can be computed from any climate variable and are combined with `Threshold` objects
    to create personalized indicators.

    Parameters
    ----------
    name: str
        The name of the indicator.
    process: Callable[..., DataArray]
        The function that processes the indicator.
    definition: str
        The definition of the indicator.
    check_vars: Callable[[list[ClimateVariable], GenericIndicator], None], optional
        A function that checks if the variables meet the indicator requirements.
        Defaults to None.
    sampling_methods: list[str], optional
        A list of sampling methods that can be used with the indicator.
        Defaults to None.
    missing: str, optional
        The method for handling missing values. Defaults to "any".
    missing_options: dict, optional
        Additional options for handling missing values. Defaults to None.
    qualifiers: tuple, optional
        Additional qualifiers for the indicator. Defaults to ().


    Attributes
    ----------
    missing: str
        The method for handling missing values.
    missing_options: dict | None
        Additional options for handling missing values.
    """  # noqa: E501

    missing: str
    missing_options: dict | None

    def __init__(
        self,
        name: str,
        process: Callable[..., DataArray],
        definition: str,
        check_vars: (
            Callable[[list[ClimateVariable], GenericIndicator], None] | None
        ) = None,
        sampling_methods: list[str] | None = None,
        missing: str = "any",
        missing_options: dict | None = None,
        qualifiers: tuple = (),
    ) -> None:
        """
        Initialize a GenericIndicator object.

        Parameters
        ----------
        name : str
            The name of the indicator.
        process : Callable[..., DataArray]
            The processing function of the indicator.
        definition : str
            A definition for the indicator.
        check_vars : Callable[[list[ClimateVariable], GenericIndicator], None] | None, optional
            A function that checks the variables used by the indicator, by default None.
        sampling_methods : list[str] | None, optional
            The sampling methods used by the indicator, by default None.
        missing : str, optional
            The method for handling missing values, by default "any".
        missing_options : Any, optional
            The options for handling missing values, by default None.
        qualifiers : tuple, optional
            The qualifiers for the indicator, by default ().

        Raises
        ------
        ValueError
            If `missing_options` is set with `missing` method being from context.

        Notes
        -----
        See the `GenericIndicatorRegistry` class for a list of available indicators.

        Examples
        --------
        >>> from icclim.generic_indices import GenericIndicator
        >>> def process(climate_vars, resample_freq):
        ...     out = climate_vars[0].studied_data + climate_vars[1].studied_data
        ...     out.resample(time=resample_freq).mean()
        ...     return out
        >>> def check_vars(climate_vars, indicator):
        ...     if len(climate_vars) != 2:
        ...         raise ValueError(
        ...             "This indicator requires exactly 2 climate variables."
        ...         )
        >>> indicator = GenericIndicator(
        ...     name="test",
        ...     process=process,
        ...     definition="This is a test indicator",
        ...     check_vars=check_vars,
        ...     sampling_methods=["daily"],
        ...     missing="skip",
        ...     missing_options=None,
        ...     qualifiers=(),
        ... )

        """  # noqa: E501
        super().__init__()
        self.missing_options = missing_options

        # assign missing method name; default is "any"
        self.missing = missing  # <-- make sure to assign this first

        # Assign the missing method (callable) directly
        self._missing = MISSING_METHODS[self.missing]
        if self.missing_options:
            missing_method.validate(**self.missing_options)
        en_indicator_templates = deepcopy(INDICATORS_TEMPLATES_EN[name])
        self.name = name
        self.process = process
        self.standard_name = en_indicator_templates["standard_name"]
        self.cell_methods = en_indicator_templates["cell_methods"]
        self.long_name = en_indicator_templates["long_name"]
        self.check_vars = check_vars
        self.definition = definition
        self.qualifiers = qualifiers
        self.sampling_methods = (
            sampling_methods if sampling_methods is not None else [RESAMPLE_METHOD]
        )

    def preprocess(
        self,
        climate_vars: list[ClimateVariable],
        jinja_scope: dict[str, Any],
        output_frequency: Frequency,
        src_freq: Frequency,
        output_unit: str | None,
        coef: float | None,
        sampling_method: str,
    ) -> list[ClimateVariable]:
        """
        Preprocesses the climate variables before computing the indicator.

        Parameters
        ----------
        climate_vars : list[ClimateVariable]
            The list of climate variables to be preprocessed.
        jinja_scope : dict[str, Any]
            The Jinja scope used for formatting the template.
        output_frequency : Frequency
            The desired frequency of the output.
        src_freq : Frequency
            The source frequency of the climate variables.
        output_unit : str | None
            The desired output unit of the indicator. If None, no unit conversion is
            performed.
        coef : float | None
            The coefficient to multiply the climate variable data with. If None,
            no multiplication is performed.
        sampling_method : str
            The sampling method used for some specific indicators.
            See `difference_of_means` for example.

        Returns
        -------
        list[ClimateVariable]
            The preprocessed climate variables.

        """
        self._check_for_invalid_setup(climate_vars, sampling_method)

        # >>> PATCHED: Convert thresholds to Kelvin if in Celsius
        for cv in climate_vars:
            if hasattr(cv, "threshold") and cv.threshold is not None:
                if cv.threshold.unit in ["C", "degC", "degree_Celsius"]:
                    # Make a copy, update internal fields
                    new_threshold = deepcopy(cv.threshold)
                    new_threshold._value = new_threshold.value + 273.15
                    new_threshold._unit = "K"
                    cv.threshold = new_threshold

        if output_unit is not None:
            if _is_amount_unit(output_unit):
                climate_vars = _convert_rates_to_amounts(
                    climate_vars=climate_vars,
                    output_unit=output_unit,
                )
            elif _is_a_diff_indicator(self) and output_unit != "%":
                # [gh:255] Indicators computing the difference between two
                # variables must first convert the units of input variables
                # to the expected output unit in order to avoid converting
                # the output of the difference.
                # This is because a difference of relative units is not equivalent
                # to a difference of absolute on scale units.
                # In other words: a 15 Kelvin difference *is* equivalent
                # to a 15 degC difference, but if we would convert the unit after
                # computing the difference, we could get -258.15 degC from the
                # 15 Kelvin.
                for climate_var in climate_vars:
                    climate_var.studied_data = convert_units_to(
                        climate_var.studied_data,
                        target=output_unit,
                    )
        if coef is not None:
            for climate_var in climate_vars:
                climate_var.studied_data = coef * climate_var.studied_data
        if output_frequency.indexer:
            for climate_var in climate_vars:
                climate_var.studied_data = select_time(
                    climate_var.studied_data,
                    **output_frequency.indexer,
                    drop=True,
                )
        _check_data(climate_vars, src_freq.pandas_freq)
        _check_cf(climate_vars)
        self._format_template(jinja_scope=jinja_scope)
        return climate_vars

    def postprocess(
        self,
        result: DataArray,
        climate_vars: list[ClimateVariable],
        output_freq: str,
        src_freq: str,
        indexer: dict,
        out_unit: str | None,
    ) -> DataArray:
        """
        Postprocesses the result of the indicator computation.

        Parameters
        ----------
        result : DataArray
            The result of the indicator computation.
        climate_vars : list[ClimateVariable]
            The list of climate variables used for the computation.
        output_freq : str
            The desired output frequency of the postprocessed result.
        src_freq : str
            The source frequency of the input data.
        indexer : dict
            The indexer used to subset the input data.
        out_unit : str | None
            The desired output unit of the postprocessed result.
            If None, no unit conversion is performed.

        Returns
        -------
        DataArray
            The postprocessed result.
        """
        """
        >>> PATCHED: Difference-aware postprocess
        Convert absolute temperatures to degC at the very end.
        Temperature differences (deltaT) remain in K (numerically identical to degC differences).
        Precipitation / amounts handled normally using hydro context.
        """

        if out_unit in ["C", "degC", "degree_Celsius"]:
            # >>> DIFF-AWARE: only convert absolute temps
            if not _is_a_diff_indicator(self):
                result.values = result.values - 273.15
                result.attrs["units"] = "degC"
            else:
                # Differences: keep in K, label as degC
                result.attrs["units"] = "degC"

        elif out_unit is not None and _is_amount_unit(out_unit):
            # Use Pint dimensionality + hydro context for precipitation-like units
            current_unit = result.attrs.get("units", None)
            if current_unit is not None:
                try:
                    q = 1 * ureg(current_unit)
                    if q.check("[mass] / [length] ** 2 / [time]") or q.check(
                        "[mass] / [length] ** 2"
                    ):
                        # Always use hydro context for rates and amounts
                        with context_hydro:
                            result = convert_units_to(result, out_unit, context="hydro")
                    else:
                        result = convert_units_to(result, out_unit)
                except Exception as e:
                    print(
                        f"_convert_rates_to_amounts: exception {e} for unit '{current_unit}', skipping"
                    )
            else:
                result = convert_units_to(result, out_unit)

        if self.missing != "skip" and indexer is not None:
            # reference variable is a subset of the studied variable,
            # so no need to check it.
            das = filter(lambda cv: not cv.is_reference, climate_vars)
            das = (cv.studied_data for cv in das)
            das = list(das)
            if "time" in result.dims:
                # If src_freq cannot be inferred by xclim, fall back to universal check_freq
                if src_freq is None:
                    try:
                        from icclim._core.generic.functions import check_freq

                        src_freq = check_freq(result, dim="time")
                    except Exception:
                        src_freq = "D"  # safe fallback: daily

                result = self._handle_missing_values(
                    in_data=das,
                    out_data=result,
                    resample_freq=output_freq,
                    src_freq=src_freq,
                    indexer=indexer,
                )

        for prop in self.templated_properties:
            result.attrs[prop] = getattr(self, prop)
        result.attrs["history"] = ""
        return result

    # >>> PATCHED helper: difference-aware flag
    def _is_a_diff_indicator(indicator: Indicator) -> bool:
        return "compute_diff" in indicator.qualifiers

    def __call__(self, config: IndexConfig) -> DataArray:
        """
        Compute the indicator based on the given configuration.

        Parameters
        ----------
        config : IndexConfig
            The configuration object containing the settings for computing the
            indicator.

        Returns
        -------
        DataArray
            The computed indicator as a DataArray.
        """
        src_freq = config.climate_variables[0].source_frequency
        base_jinja_scope = {
            "np": np,
            "enumerate": enumerate,
            "len": len,
            "output_freq": config.frequency,
            "source_freq": src_freq,
        }
        climate_vars_meta = _get_climate_vars_metadata(
            config.climate_variables,
            src_freq,
            base_jinja_scope,
            jinja_env,
        )
        jinja_scope: dict[str, Any] = {
            "min_spell_length": config.min_spell_length,
            "rolling_window_width": config.rolling_window_width,
            "climate_vars": climate_vars_meta,
            "is_compared_to_reference": config.is_compared_to_reference,
            "reference_period": config.reference_period,
        }
        jinja_scope.update(base_jinja_scope)
        climate_vars = self.preprocess(
            climate_vars=config.climate_variables,
            jinja_scope=jinja_scope,
            output_frequency=config.frequency,
            src_freq=src_freq,
            output_unit=config.out_unit,
            coef=config.coef,
            sampling_method=config.sampling_method,
        )
        result = self.process(
            climate_vars=climate_vars,
            resample_freq=config.frequency,
            min_spell_length=config.min_spell_length,
            rolling_window_width=config.rolling_window_width,
            group_by_freq=config.frequency.group_by_key,
            is_compared_to_reference=config.is_compared_to_reference,
            logical_link=config.logical_link,
            date_event=config.date_event,
            source_freq_delta=src_freq.delta,
            to_percent=config.out_unit == "%",
            sampling_method=config.sampling_method,
        )
        return self.postprocess(
            result,
            climate_vars=climate_vars,
            output_freq=config.frequency.pandas_freq,
            src_freq=src_freq.pandas_freq,
            indexer=config.frequency.indexer,
            out_unit=config.out_unit,
        )

    def __eq__(self, other: object) -> bool:
        """
        Check if two GenericIndicator objects are equal.

        Parameters
        ----------
        other : Any
            The object to compare with.

        Returns
        -------
        bool
            True if the two objects are equal, False otherwise.
        """
        return (
            isinstance(other, GenericIndicator)
            and self.long_name == other.long_name
            and self.standard_name == other.standard_name
            and self.process == other.process
        )

    def __str__(self) -> str:
        """
        Return the name of the indicator.

        Returns
        -------
        str
            The name of the indicator.
        """
        return self.name

    def _check_for_invalid_setup(
        self,
        climate_vars: list[ClimateVariable],
        sampling_method: str,
    ) -> None:
        if not _same_freq_for_all(climate_vars):
            msg = (
                "All variables must have the same time frequency (for example daily) to"
                " be compared with each others, but this was not the case."
            )
            raise InvalidIcclimArgumentError(msg)
        if sampling_method not in self.sampling_methods:
            msg = (
                f"{self.name} can only be computed with the following"
                f" sampling_method(s): {self.sampling_methods}"
            )
            raise InvalidIcclimArgumentError(msg)
        if self.check_vars is not None:
            # Run indicator specific check method
            self.check_vars(climate_vars, self)

    def _format_template(self, jinja_scope: dict) -> None:
        for templated_property in self.templated_properties:
            template = jinja_env.from_string(
                getattr(self, templated_property),
                globals=jinja_scope,
            )
            setattr(self, templated_property, template.render())

    def _handle_missing_values(
        self, in_data, out_data, resample_freq=None, src_freq=None, indexer=None
    ):
        """
        Handle missing values in climate index computations.

        Parameters
        ----------
        in_data : list[xr.DataArray]
            Input DataArrays with time coordinates.
        out_data : xr.DataArray
            Output DataArray from the index computation.
        resample_freq : str, optional
            Target resampling frequency (e.g. "M", "Y").
        src_freq : str, optional
            Source timestep frequency (e.g. "D").
        indexer : dict, optional
            Extra arguments used by some missing value methods.
        """
        import numpy as np
        from xarray import DataArray
        from xclim.core.missing import MISSING_METHODS

        missing_class = MISSING_METHODS[self.missing]  # Get the class
        missing_obj = missing_class()  # Instantiate with no args

        # We flag periods according to the missing method. Skip variables without a time coordinate.
        miss = (
            missing_obj(
                da, freq=resample_freq, src_timestep=src_freq, **(indexer or {})
            )
            for da in in_data
            if "time" in da.coords
        )

        # Reduce by logical OR across all input masks
        mask = reduce(np.logical_or, miss)

        # Reindex mask to match output if needed
        if isinstance(mask, DataArray) and mask.time.size < out_data.time.size:
            mask = mask.reindex(time=out_data.time, fill_value=True)

        return out_data.where(~mask)


def _same_freq_for_all(climate_vars: list[ClimateVariable]) -> bool:
    if len(climate_vars) == 1:
        return True
    freqs = [xr.infer_freq(a.studied_data.time) for a in climate_vars]
    return all(x == freqs[0] for x in freqs[1:])


def _get_climate_vars_metadata(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    jinja_scope: dict[str, Any],
    jinja_env: jinja2.Environment,
) -> list[dict[str, str]]:
    return [
        c_var.build_indicator_metadata(
            resample_freq,
            must_run_bootstrap(c_var.studied_data, c_var.threshold),
            jinja_scope,
            jinja_env,
        )
        for c_var in climate_vars
    ]


def _convert_rates_to_amounts(
    climate_vars: list[ClimateVariable], output_unit: str
) -> list[ClimateVariable]:
    """
    Convert rate-like climate variables to amount units using xclim's rate2amount.
    Handles both classic rates (e.g., mm/s) and precipitation (kg m-2 / time)
    using a dedicated Pint context.
    """
    for climate_var in climate_vars:
        # Get the current unit of the variable
        current_unit = climate_var.studied_data.attrs.get("units", None)
        if current_unit is None:
            continue

        # Skip conversion if already an amount
        if _is_amount_unit(current_unit):
            continue

        try:
            # Clean unit string for Pint parsing
            unit_clean = current_unit.replace(" ", "*").replace("-", "**")
            q = 1 * ureg(unit_clean)
            dims = q.dimensionality

            # Determine if this is a precipitation-like unit (mass/area or mass/area/time)
            is_precip = (
                dims == ureg("kg / m^2").dimensionality
                or dims == ureg("kg / m^2 / s").dimensionality
                or "kg/m2" in current_unit.replace(" ", "").lower()
            )

            if is_precip:
                # >>> Hydro context applied here
                print(
                    f"Converting {climate_var.name}: {current_unit} -> {output_unit} with hydro context"
                )
                with context_hydro:
                    da = rate2amount(climate_var.studied_data, out_units=output_unit)
            else:
                # Non-precipitation rates: normal conversion
                da = rate2amount(climate_var.studied_data, out_units=output_unit)

            # Update the variable with converted data
            climate_var.studied_data = da.astype("float64")

        except Exception as e:
            # Skip on error but report it
            print(
                f"_convert_rates_to_amounts: exception {e} for unit '{current_unit}', skipping"
            )
            continue

    return climate_vars


def _is_amount_unit(unit: str) -> bool:
    """
    Return True if `unit` is an amount unit (length-like), False otherwise.
    """
    unit = unit.strip()
    offset_units = ["K", "degC", "degree_Celsius", "C", "?C"]
    if any(u in unit for u in offset_units):
        return False
    try:
        # Fix for Pint parsing: replace spaces with * and '-' with '**'
        unit_clean = unit.replace(" ", "*").replace("-", "**")
        q = 1 * ureg(unit_clean)
        # Compare to the dimensionality of 1 meter
        return q.dimensionality == (1 * ureg("m")).dimensionality
    except Exception as e:
        print(f"_is_amount_unit: exception {e} for unit '{unit}', returning False")
        return False


def _check_cf(climate_vars: list[ClimateVariable]) -> None:
    """Compare metadata attributes to CF-Convention standards.

    Default cfchecks use the specifications in `xclim.core.utils.VARIABLES`,
    assuming the indicator's inputs are using the CMIP6/xclim variable names
    correctly.
    Variables absent from these default specs are silently ignored.

    When subclassing this method, use functions decorated using
    `xclim.core.options.cfcheck`.
    """
    for da in climate_vars:
        with contextlib.suppress(KeyError):
            # Silently ignore unknown variables.
            cfcheck_from_name(str(da.name), da.studied_data)


def _check_data(climate_vars: list, src_freq: str) -> None:
    if src_freq is None:
        return
    for climate_var in climate_vars:
        da = climate_var.studied_data
        if "time" in da.coords and da.time.ndim == 1 and len(da.time) > 3:
            inferred_freq = check_freq(da, dim="time", strict=True)
            if inferred_freq != src_freq:
                raise InvalidIcclimArgumentError(
                    f"[icclim] Frequency mismatch for variable '{climate_var.name}': "
                    f"expected '{src_freq}', inferred '{inferred_freq}'"
                )


def _is_a_diff_indicator(indicator: Indicator) -> bool:
    return "compute_diff" in indicator.qualifiers



"""
Define the StandardVariable class and the StandardVariableRegistry class.

The StandardVariable class represents a standard variable used in climate data analysis.
The StandardVariableRegistry class is a registry that stores instances of
`StandardVariable`.

StandardVariable are mainly used to compute the metadata of the output of the generic
indices.
"""

from __future__ import annotations

import dataclasses
from collections.abc import Hashable

from icclim._core.constants import PART_OF_A_WHOLE_UNIT
from icclim._core.model.registry import Registry


@dataclasses.dataclass
class StandardVariable(Hashable):
    """
    StandardVariable represents a typical variable used in climate data analysis.

    Parameters
    ----------
    short_name : str
        The short name of the variable.
    standard_name : str
        The standard name of the variable.
    long_name : str
        The long name of the variable.
    aliases : list[str]
        A list of aliases for the variable.
    default_units : str
        The default units of the variable.
    """

    short_name: str
    standard_name: str
    long_name: str
    aliases: list[str]
    default_units: str

    def __hash__(self) -> int:
        """Build the hash of the StandardVariable object."""
        return hash(self.short_name + self.standard_name)

    def get_metadata(self) -> dict[str, str]:
        """Get the metadata of the StandardVariable object."""
        return {
            "standard_name": self.standard_name,
            "long_name": self.long_name,
            "short_name": self.short_name,
        }


class StandardVariableRegistry(Registry[StandardVariable]):
    """StandardVariableRegistry stores instances of StandardVariable such as PR, TAS."""

    _item_class = StandardVariable
    PR = StandardVariable(
        short_name="pr",
        standard_name="precipitation_flux",
        long_name="precipitation",
        aliases=[
            "pr",
            "prAdjust",
            "prec",
            "rr",
            "precip",
            "RR",
            "PRECIP",
            "Precip",
        ],
        default_units="mm day-1",
    )
    TAS = StandardVariable(
        short_name="tg",
        standard_name="average_air_temperature",
        long_name="average air temperature",
        aliases=[
            "tas",
            "tavg",
            "ta",
            "tasAdjust",
            "tmean",
            "tm",
            "tg",
            "meant",
            "TMEAN",
            "Tmean",
            "TM",
            "TG",
            "MEANT",
            "meanT",
            "tasmidpoint",
        ],
        default_units="degree_Celsius",
    )
    TAS_MIN = StandardVariable(
        short_name="tn",
        standard_name="minimum_air_temperature",  # not CF nor CMIP (air_temperature)
        long_name="minimum air temperature",
        aliases=[
            "tasmin",
            "tasminAdjust",
            "tmin",
            "tn",
            "mint",
            "TMIN",
            "Tmin",
            "TN",
            "MINT",
            "minT",
        ],
        default_units="degree_Celsius",
    )
    TAS_MAX = StandardVariable(
        short_name="tx",
        standard_name="maximum_air_temperature",  # not CF nor CMIP (air_temperature)
        long_name="maximum air temperature",
        aliases=[
            "tasmax",
            "tasmaxAdjust",
            "tmax",
            "tx",
            "maxt",
            "TMAX",
            "Tmax",
            "TX",
            "MAXT",
            "maxT",
        ],
        default_units="degree_Celsius",
    )
    HURS = StandardVariable(
        short_name="hurs",
        standard_name="relative_humidity",
        long_name="relative humidity",
        aliases=["hurs", "hursAdjust", "rh", "RH"],
        default_units="1",  # %
    )
    PS = StandardVariable(
        short_name="ps",
        standard_name="surface_air_pressure ",
        long_name="Surface Air Pressure",
        aliases=["ps"],
        default_units="Pa",
    )
    PSL = StandardVariable(
        short_name="psl",
        standard_name="air_pressure_at_sea_level ",
        long_name="Sea Level Pressure",
        aliases=["psl", "mslp", "slp", "pp", "MSLP", "SLP", "PP"],
        default_units="Pa",
    )
    SND = StandardVariable(
        short_name="snd",
        standard_name="surface_snow_thickness",
        long_name="Snow Depth",
        aliases=["snd", "sd", "SD"],
        default_units="m",
    )
    SUND = StandardVariable(
        short_name="sund",
        standard_name="duration_of_sunshine",
        long_name="duration of sunshine",
        aliases=["sund", "ss", "SS"],
        default_units="s",
    )
    WSGS_MAX = StandardVariable(
        short_name="wsgs_max",
        standard_name="wind_speed_of_gust",
        long_name="wind speed of gust",
        aliases=["wsgsmax", "wsgs_max", "fx", "FX"],
        default_units="m s-1",
    )
    SFC_WIND = StandardVariable(
        short_name="sfcWind",
        standard_name="wind_speed",
        long_name="Near-Surface Wind Speed",
        aliases=["sfcWind", "sfcwind", "sfc_wind", "fg", "FG"],
        default_units="m s-1",
    )
    SNW = StandardVariable(
        short_name="snw",
        standard_name="surface_snow_amount",
        long_name="surface snow amount",
        aliases=["snw", "swe", "SW"],
        default_units="kg m-2",
    )
    EVSPSBL = StandardVariable(
        short_name="evspsbl",
        standard_name="water_evapotranspiration_flux",
        long_name="Evaporation Including Sublimation and Transpiration",
        aliases=["evspsbl", "water_evaporation_flux"],
        default_units="kg m-2 s-1",
    )
    HUSS = StandardVariable(
        short_name="huss",
        standard_name="specific_humidity",
        long_name="Near-Surface Specific Humidity",
        aliases=["huss"],
        default_units=PART_OF_A_WHOLE_UNIT,
    )
    UAS = StandardVariable(
        short_name="uas",
        standard_name="eastward_wind",
        long_name="Eastward Near-Surface Wind",
        aliases=["uas"],
        default_units="m s-1",
    )
    VAS = StandardVariable(
        short_name="vas",
        standard_name="northward_wind",
        long_name="Northward Near-Surface Wind",
        aliases=["vas"],
        default_units="m s-1",
    )
    CLT = StandardVariable(
        short_name="clt",
        standard_name="cloud_area_fraction",
        long_name="Total Cloud Cover Percentage",
        aliases=["clt"],
        default_units="%",
    )
    RSDS = StandardVariable(
        short_name="rsds",
        standard_name="surface_downwelling_shortwave_flux_in_air",
        long_name="Surface Downwelling Shortwave Radiation",
        aliases=["rsds", "surface_downwelling_shortwave_flux"],
        default_units="W m-2",
    )
    RLDS = StandardVariable(
        short_name="rlds",
        standard_name="surface_downwelling_longwave_flux_in_air",
        long_name="Surface Downwelling Longwave Radiation",
        aliases=["rlds", "surface_downwelling_longwave_flux"],
        default_units="W m-2",
    )
    RSUS = StandardVariable(
        short_name="rsds",
        standard_name="surface_upwelling_shortwave_flux_in_air",
        long_name="Surface Upwelling Shortwave Radiation",
        aliases=["rsus", "surface_upwelling_shortwave_flux"],
        default_units="W m-2",
    )
    RLUS = StandardVariable(
        short_name="rlds",
        standard_name="surface_upwelling_longwave_flux_in_air",
        long_name="Surface Upwelling Longwave Radiation",
        aliases=["rlus", "surface_upwelling_longwave_flux"],
        default_units="W m-2",
    )
    OROG = StandardVariable(
        short_name="orog",
        standard_name="surface_altitude",
        long_name="Surface Altitude",
        aliases=["orog"],
        default_units="m",
    )
    SFTLF = StandardVariable(
        short_name="sftlf",
        standard_name="land_area_fraction",
        long_name="Percentage of the Grid Cell Occupied by Land",
        aliases=["sftlf"],
        default_units="%",
    )
    WIND_TO_DIRECTION = StandardVariable(
        short_name="DD",
        standard_name="wind_to_direction",
        long_name="Daily mean wind direction",
        aliases=["dd"],
        default_units="degree",
    )
    # TODO @bzah: add tier1 and tier2 aliases from cmip6/cordex
    # https://docs.google.com/spreadsheets/d/1qUauozwXkq7r1g-L4ALMIkCNINIhhCPx/edit?rtpof=true&sd=true#gid=1672965248
    # https://github.com/cerfacs-globc/icclim/issues/289

    @staticmethod
    def get_item_aliases(item: StandardVariable) -> list[str]:
        """
        Return a list of aliases for the given StandardVariable.

        Parameters
        ----------
        item : StandardVariable
            The StandardVariable object for which to retrieve the aliases.

        Returns
        -------
        list[str]
            A list of aliases for the given StandardVariable.
        """
        aliases = list(map(str.upper, item.aliases))
        aliases.append(item.standard_name.upper())
        aliases.append(item.long_name.upper())
        return aliases



"""
Contain the classes for quantile interpolation.

QuantileInterpolation class and the QuantileInterpolationRegistry class
are defined here.
"""

from __future__ import annotations

import dataclasses

from icclim._core.model.registry import Registry


@dataclasses.dataclass
class QuantileInterpolation:
    """
    Class for performing quantile interpolation.

    Parameters
    ----------
    name : str
        The name of the interpolation method.
    alpha : float
        The alpha parameter for the interpolation.
    beta : float
        The beta parameter for the interpolation.
    """

    name: str
    alpha: float
    beta: float


class QuantileInterpolationRegistry(Registry[QuantileInterpolation]):
    """
    Registry of quantile interpolation methods.

    Only 2 methods are available: LINEAR and MEDIAN_UNBIASED.
    """

    _item_class = QuantileInterpolation

    LINEAR = QuantileInterpolation("linear", 1, 1)
    MEDIAN_UNBIASED = QuantileInterpolation("median_unbiased", 1.0 / 3, 1.0 / 3)



"""
User indices module for icclim.

User indices are a way of creating custom climate indices since icclim v4.
It allows to define a dictionary to describe the computation of a climate index.
Since icclim v6 it is reccomended to use generic indices instead, as they are more
flexible and have a better API.
See :ref:`generic_indices_recipes` for more information on generic indices.

User indices are not deprecated but are not recommended for new code.
"""



"""
BoundedThreshold module.

A `BoundedThreshold` is a composite threshold that binds two other thresholds
for example "> 95 doy_per AND >= 30 deg_C".
"""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from icclim._core.generic.threshold.threshold_templates import (
    EN_THRESHOLD_TEMPLATE,
    ThresholdMetadata,
)
from icclim._core.model.threshold import (
    Threshold,
    ThresholdBuilderInput,
)
from icclim.exception import InvalidIcclimArgumentError

if TYPE_CHECKING:
    from collections.abc import Sequence

    import jinja2
    import xarray as xr
    from xarray import DataArray

    from icclim._core.model.logical_link import LogicalLink


class BoundedThreshold(Threshold):
    """
    Threshold that binds two other thresholds (e.g. "> 95 doy_per AND >= 30 deg_C").

    The logical link must be either "and" or "or".
    """

    left_threshold: Threshold
    right_threshold: Threshold
    logical_link: LogicalLink

    @property
    def unit(self) -> str | None:
        """
        The unit of the bounded threshold.

        Returns
        -------
        str | None
            The unit of the threshold if both thresholds have the same unit,
            otherwise None.
        """
        if self.left_threshold.unit == self.right_threshold.unit:
            return self.left_threshold.unit
        return None

    @unit.setter
    def unit(self, unit: str) -> None:
        """
        Set the unit on each of the thresholds.

        Parameters
        ----------
        unit : str
            The unit to set.

        Notes
        -----
        This setter will attempt a unit conversion using units found in xclim's pint
        registry.
        """
        self.left_threshold.unit = unit
        self.right_threshold.unit = unit

    def __init__(
        self,
        thresholds: Sequence[Threshold | str | ThresholdBuilderInput],
        logical_link: LogicalLink,
        initial_query: str | None,
        **kwargs,  # noqa: ARG002
    ) -> None:
        if len(thresholds) != 2:
            msg = (
                f"BoundedThreshold can only be built on 2 thresholds, {len(thresholds)}"
                f" were found."
            )
            raise InvalidIcclimArgumentError(msg)
        self.left_threshold = self._build_thresh(thresholds[0])
        self.right_threshold = self._build_thresh(thresholds[1])
        if self.left_threshold == self.right_threshold:
            msg = (
                f"BoundedThreshold must be built on 2 **different** thresholds, here"
                f" both were {self.left_threshold.initial_query}"
            )
            raise InvalidIcclimArgumentError(msg)
        self.logical_link = logical_link
        self.initial_query = initial_query

    def compute(
        self,
        comparison_data: xr.DataArray,
        override_op: Callable[[DataArray, DataArray], DataArray] | None = None,
        **kwargs,
    ) -> DataArray:
        """
        Compute the threshold exceedance value.

        Uses the logical link to combine the result of both thresholds.

        Parameters
        ----------
        comparison_data : xr.DataArray
            The data array to compare with the threshold value.
        override_op : Callable[[DataArray, DataArray], DataArray] | None, optional
            A custom override function to compute the threshold value.
            If provided, this function will be used instead of the default operator.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        DataArray
            The computed threshold value.

        Notes
        -----
        If `override_op` is not None, the `override_op` function will be used to
        compute the thresholds exceedances on both thresholds.
        """
        left_res = self.left_threshold.compute(
            comparison_data,
            override_op=override_op,
            **kwargs,
        )
        right_res = self.right_threshold.compute(
            comparison_data,
            override_op=override_op,
            **kwargs,
        )
        return self.logical_link.compute([left_res, right_res])

    def format_metadata(
        self,
        *,
        jinja_scope: dict[str, Any],
        jinja_env: jinja2.Environment,
        **kwargs,
    ) -> ThresholdMetadata:
        """
        Generate the metadata for the bounded threshold.

        These metadata are used to fill the generic template.

        Parameters
        ----------
        jinja_scope : dict
            The jinja scope, it contains the variables to be used in the jinja template.
        jinja_env : jinja2.Environment
            The jinja environment, it contains the jinja rendering engine.
        **kwargs
            Additional keyword arguments, ignored for compatibility with other
            `format_metadata` methods.

        Returns
        -------
        ThresholdMetadata
            The metadata for the threshold.

        """
        templates = self._get_metadata_templates()
        conf = {
            "left_threshold": self.left_threshold.format_metadata(
                jinja_scope=jinja_scope,
                jinja_env=jinja_env,
                **kwargs,
            ),
            "logical_link": self.logical_link,
            "right_threshold": self.right_threshold.format_metadata(
                jinja_scope=jinja_scope,
                jinja_env=jinja_env,
                **kwargs,
            ),
        }
        conf.update(jinja_scope)
        return {
            k: jinja_env.from_string(v, globals=conf).render()
            for k, v in templates.items()
        }

    def __eq__(self, other: BoundedThreshold) -> bool:
        """
        Check if the comparison threshold is equivalent to `self`.

        BoundedThreshold equality is based on the reflexive property of logical links (
        and/or).
        Thus, the left_threshold of `self` can be either equal to the
        left_threshold or the right_threshold of `other`.
        Same goes for right_threshold of self.

        Parameters
        ----------
        other : BoundedThreshold
            The comparison threshold.

        Returns
        -------
            True if the the comparison threshold `other` is equivalent to `self`.
        """
        return (
            isinstance(other, BoundedThreshold)
            and self.initial_query == other.initial_query
            and (self.left_threshold in (other.left_threshold, other.right_threshold))
            and (self.right_threshold in (other.left_threshold, other.right_threshold))
            and self.logical_link == other.logical_link
        )

    def _build_thresh(
        self,
        thresh_input: Threshold | str | ThresholdBuilderInput,
    ) -> Threshold:
        from icclim.threshold.factory import build_threshold

        if isinstance(thresh_input, Threshold):
            return thresh_input
        if isinstance(thresh_input, str):
            return build_threshold(thresh_input)
        return build_threshold(**thresh_input)

    def _get_metadata_templates(self) -> ThresholdMetadata:
        return EN_THRESHOLD_TEMPLATE["bounded_threshold"]



"""Global metadata model to be added to the output netCDF file."""

from __future__ import annotations

from typing import TypedDict


class GlobalMetadata(TypedDict):
    """Global metadata model.

    Attributes
    ----------
    history : str or None
        The CF history attribute.
    source : str or None
        The source of the data.
    time_encoding : dict or None
        The time encoding information to be read from ds.time.encoding.
    """

    history: str | None
    source: str | None
    time_encoding: dict | None



"""Contain the CalcOperationLike and CalcOperation classes."""

from __future__ import annotations

import dataclasses
from collections.abc import Hashable
from typing import Literal

from icclim._core.model.registry import Registry

CalcOperationLike = Literal[
    "max",
    "min",
    "sum",
    "mean",
    "nb_events",
    "max_nb_consecutive_events",
    "run_mean",
    "run_sum",
    "anomaly",
]


@dataclasses.dataclass
class CalcOperation(Hashable):
    """Represent a calculation operation for a user index."""

    name: str

    def __hash__(self) -> int:
        """Return the hash of the CalcOperation using its name."""
        return hash(self.name)


class CalcOperationRegistry(Registry[CalcOperation]):
    """Registry for CalcOperation instances."""

    _item_class = CalcOperation

    MAX = CalcOperation("max")
    MIN = CalcOperation("min")
    SUM = CalcOperation("sum")
    MEAN = CalcOperation("mean")
    EVENT_COUNT = CalcOperation("nb_events")
    MAX_NUMBER_OF_CONSECUTIVE_EVENTS = CalcOperation(
        "max_nb_consecutive_events",
    )
    RUN_MEAN = CalcOperation("run_mean")
    RUN_SUM = CalcOperation("run_sum")
    ANOMALY = CalcOperation("anomaly")



"""Python library for climate indices calculation."""

from icclim import dcsc, ecad, generic
from icclim._generated._ecad import *  # noqa: F403 (add ECAD api to icclim namespace)
from icclim._generated._generic import *  # noqa: F403 (add generic api to icclim namespace)
from icclim.main import index, indice, indices
from icclim.rechunk import create_optimized_zarr_store
from icclim.threshold.factory import build_threshold

__all__ = [
    # -- Threshold factory function
    "build_threshold",
    # -- Rechunk function
    "create_optimized_zarr_store",
    # -- Base functions
    "dcsc",
    "ecad",
    "generic",
    "index",
    "indice",  # deprecated
    "indices",
]

__version__ = "7.0.4"
